module gcbpv where

open import Data.Empty
open import Data.Unit
open import Data.Sum hiding (map)
open import Relation.Binary.PropositionalEquality
open import Data.Nat
open import Data.Fin
open import Data.Vec
import Data.Vec
import Data.Vec.Properties
open ≡-Reasoning

lookup-map : ∀ {a b n} {A : Set a} {B : Set b}
             (i : Fin n) (f : A → B) (xs : Vec A n) →
             lookup i (map f xs) ≡ f (lookup i xs)
lookup-map zero    f (x ∷ xs) = refl
lookup-map (suc i) f (x ∷ xs) = lookup-map i f xs

data Stoup : Set where
  None : Stoup
  One  : Stoup

-- | Semantics
-- | ⟦ S ⟧st X =~ ⟦ S ⟧ -> X
⟦_⟧st : Stoup -> Set -> Set
⟦ None ⟧st X = ⊤
⟦ One ⟧st X = X

record StCtx (X : Set) : Set where
  constructor stctx
  field
    S : Stoup
    ty : ⟦ S ⟧st X

cdot : forall (X : Set) -> StCtx X
cdot X = record { S = None ; ty = tt }

hstoup : forall {Y : Set} -> Y -> StCtx Y
hstoup B = record { S = One ; ty = B }

data Dimension : Set where
  Top : Dimension
  Bot : Dimension

DCtx : Set
DCtx = StCtx ⊤

onevar : StCtx ⊤
onevar = hstoup tt

infixl 90 _⟨_⟩v
infixl 90 _⟨_⟩vsq
infixl 90 _⟨_⟩c

data VType : DCtx -> Set
_⟨_⟩v : VType onevar -> Dimension -> VType (cdot _)
record VTC : Set
data CType : DCtx -> Set
_⟨_⟩c : CType onevar -> Dimension -> CType (cdot _)

-- | Transitivvity-composable pair of value types
-- | "Value Transitivity Composite"
record VTC where
  inductive
  field
    A⊤ : VType onevar
    A⊥ : VType onevar
    eq : A⊤ ⟨ Bot ⟩v ≡ A⊥ ⟨ Top ⟩v


data VType where
  VUnit  : forall I -> VType I
  _×_    : forall {I} -> VType I -> VType I -> VType I
  U      : forall {I} -> CType I -> VType I
  vtrans : VTC -> VType onevar

VUnit _ ⟨ i ⟩v = VUnit _
(A × A') ⟨ i ⟩v = (A ⟨ i ⟩v) × (A' ⟨ i ⟩v)
U B ⟨ i ⟩v = U (B ⟨ i ⟩c)
vtrans Vsq ⟨ Top ⟩v = VTC.A⊤ Vsq ⟨ Top ⟩v -- morally this: Vsq ⟨ Top ⟩vsq ⟨ Top ⟩v
vtrans Vsq ⟨ Bot ⟩v = VTC.A⊥ Vsq ⟨ Bot ⟩v -- morally this: Vsq ⟨ Bot ⟩vsq ⟨ Bot ⟩v
-- vtrans Atop Abot p ⟨ Top ⟩v = Atop ⟨ Top ⟩v
-- vtrans Atop Abot p ⟨ Bot ⟩v = Abot ⟨ Bot ⟩v

data CType where
  _⇀_ : forall {I} -> VType I -> CType I -> CType I
  F   : forall {I} -> VType I -> CType I
  ctrans : forall (Btop : CType onevar) (Bbot : CType onevar) ->
           Btop ⟨ Bot ⟩c ≡ Bbot ⟨ Top ⟩c ->
           CType onevar

(A ⇀ B) ⟨ i ⟩c = (A ⟨ i ⟩v) ⇀ (B ⟨ i ⟩c)
F A ⟨ i ⟩c = F (A ⟨ i ⟩v)
ctrans Btop Bbot p ⟨ Top ⟩c = Btop ⟨ Top ⟩c
ctrans Btop Bbot p ⟨ Bot ⟩c = Bbot ⟨ Bot ⟩c

_⟨_⟩vsq : VTC -> Dimension -> VType onevar
Vsq ⟨ Top ⟩vsq = VTC.A⊤ Vsq
Vsq ⟨ Bot ⟩vsq = VTC.A⊥ Vsq

vrefl : VType (cdot _) -> VType onevar
crefl : CType (cdot _) -> CType onevar
vrefl (VUnit _) = VUnit _
vrefl (A × A') = vrefl A × vrefl A'
vrefl (U B) = U (crefl B)

crefl (A ⇀ B) = vrefl A ⇀ crefl B
crefl (F A) = F (vrefl A)

vreflretract : forall A i -> vrefl A ⟨ i ⟩v ≡ A
creflretract : forall B i -> crefl B ⟨ i ⟩c ≡ B

vreflretract (VUnit _) i = refl
vreflretract (A × A₁) i = cong₂ _×_ (vreflretract A i) (vreflretract A₁ i)
vreflretract (U B) i = cong U (creflretract B i)

creflretract (A ⇀ B) i = cong₂ _⇀_ (vreflretract A i) (creflretract B i)
creflretract (F A) i = cong F (vreflretract A i)

-- | For a more traditional presentation of ordering
data VLt (A A' : VType (cdot _)) : Set where
  lt : forall (Aord : VType onevar) -> Aord ⟨ Top ⟩v ≡ A -> Aord ⟨ Bot ⟩v ≡ A' -> VLt A A'
data CLt (B B' : CType (cdot _)) : Set where
  lt : forall (Bord : CType onevar) -> Bord ⟨ Top ⟩c ≡ B -> Bord ⟨ Bot ⟩c ≡ B' -> CLt B B'

-- | Some sanity checking
-- vltrefl : forall A -> VLt A A
-- vltrefl A = lt (vrefl A) (vreflretract A Top) ((vreflretract A Bot))

-- vlttrans : forall A1 A2 A3 -> VLt A1 A2 -> VLt A2 A3 -> VLt A1 A3
-- vlttrans A1 A2 A3 (lt A12 x x₁) (lt A23 x₂ x₃) = lt (vtrans A12 A23 (trans x₁ (sym x₂))) x x₃

-- cltrefl : forall B -> CLt B B
-- cltrefl B = lt (crefl B) (creflretract B Top) ((creflretract B Bot))

clttrans : forall B1 B2 B3 -> CLt B1 B2 -> CLt B2 B3 -> CLt B1 B3
clttrans B1 B2 B3 (lt B12 x x₁) (lt B23 x₂ x₃) = lt (ctrans B12 B23 (trans x₁ (sym x₂))) x x₃

-- | Now the term structure
-- Values correspond to complex values in cbpv
-- Terms correspond to both terms and complex stacks in cbpv
-- Everything is also indexed by a dimension stoup, so that we simultaneously
-- implement the term constructors and their monotonicity, meaning like the
-- types we can implement reflexivity as weakening and only have to add in
-- transitivity, beta and eta

record VCtx (I : DCtx) : Set where
  constructor vctx
  field
    Vars : ℕ
    vty  : Vec (VType I) Vars
open VCtx

infixl 20 _,,_
_,,_ : forall {I} -> VCtx I -> VType I -> VCtx I
Vars (Γ ,, A) = suc (Vars Γ)
vty  (Γ ,, A) = A ∷ vty Γ

infixl 90 _⟪_⟫v
infixl 90 _⟪_⟫vsq
infixl 90 _⟪_⟫c

_⟪_⟫v : VCtx onevar -> Dimension -> VCtx (cdot _)
Vars (Γ ⟪ i ⟫v) = Vars Γ
vty (Γ ⟪ i ⟫v)  = map (\A -> A ⟨ i ⟩v) (vty Γ)

-- ctx-app-lem : ∀ Γ i -> Γ ⟪ i ⟫v

vctxrefl : VCtx (cdot _) -> VCtx onevar
Vars (vctxrefl Γ) = Vars Γ
vty  (vctxrefl Γ) = map vrefl (vty Γ)

record VCtxTC : Set where
  field
    Vars : ℕ
    vty  : Vec VTC Vars
open VCtxTC

_⟪_⟫vsq : VCtxTC -> Dimension -> VCtx onevar
Vars (Γsq ⟪ i ⟫vsq) = Vars Γsq
vty  (Γsq ⟪ i ⟫vsq) = (map (λ Vsq → Vsq ⟨ i ⟩vsq) (vty Γsq))

vctxtrans : VCtxTC -> VCtx onevar
vctxtrans Γsq = vctx (Vars Γsq) (map vtrans (vty Γsq))

ctxttrans-dapp-lem : ∀ Γsq i -> vctxtrans Γsq ⟪ i ⟫v ≡ Γsq ⟪ i ⟫vsq ⟪ i ⟫v
ctxttrans-dapp-lem Γsq i =
  begin
    vctxtrans Γsq ⟪ i ⟫v
    ≡⟨ {!!} ⟩
    vctxtrans Γsq ⟪ i ⟫v
    ≡⟨ {!!} ⟩
    Γsq ⟪ i ⟫vsq ⟪ i ⟫v
  ∎

CCtx : DCtx -> Set
CCtx I = StCtx (CType I)

_⟪_⟫c : CCtx onevar -> Dimension -> CCtx (cdot _)
stctx None ty ⟪ i ⟫c = stctx None tt
stctx One ty ⟪ i ⟫c = stctx One (ty ⟨ i ⟩c)

cctxrefl : CCtx (cdot _) -> CCtx onevar
cctxrefl (stctx None ty) = stctx None tt
cctxrefl (stctx One ty) = stctx One (crefl ty) 

data Value : forall {I} -> VCtx I -> VType I -> Set
_⟨_⟩vv : forall {Γ : VCtx onevar} {A} ->
         Value Γ A ->
         (i : Dimension) ->
         Value (Γ ⟪ i ⟫v) (A ⟨ i ⟩v)
record ValTC : Set
data Term  : forall {I} -> VCtx I -> CCtx I -> CType I -> Set
_⟨_⟩cc : forall {Γ : VCtx onevar}{Δ}{B} ->
         Term Γ Δ B ->
         (i : Dimension) ->
         Term (Γ ⟪ i ⟫v) (Δ ⟪ i ⟫c) (B ⟨ i ⟩c)

record ValTC where
  inductive
  field
    ΓΓ : VCtxTC
    AA : VTC
    V⊤ : Value (ΓΓ ⟪ Top ⟫vsq) (AA ⟨ Top ⟩vsq)
    V⊥ : Value (ΓΓ ⟪ Bot ⟫vsq) (AA ⟨ Bot ⟩vsq)

data Value where
  var : forall {I} (Γ : VCtx I) (x : Fin (Vars Γ)) A -> A ≡ lookup x (vty Γ) -> Value Γ A
  vunit : forall {I} (Γ : VCtx I) -> Value Γ (VUnit _)
  vpair : forall {I}{Γ : VCtx I}{A}{A'} -> Value Γ A -> Value Γ A' -> Value Γ (A × A')
  vdestr : forall {I}{Γ : VCtx I}{A}{A'}{Am} ->
           Value Γ (A × A') ->
           Value (Γ ,, A ,, A') Am ->
           Value Γ Am
  vthunk : forall {I}{Γ : VCtx I}{B} ->
           Term Γ (cdot _) B ->
           Value Γ (U B)
  valtrans : forall
              (sq : ValTC) ->
              Value (vctxtrans (ValTC.ΓΓ sq)) (vtrans (ValTC.AA sq))

var Γ x A p ⟨ i ⟩vv = var (Γ ⟪ i ⟫v) x (A ⟨ i ⟩v)
  (begin
    A ⟨ i ⟩v
    ≡⟨ cong (λ A → A ⟨ i ⟩v) p ⟩
    (lookup x (vty Γ)) ⟨ i ⟩v
    ≡⟨ sym (lookup-map x (\A -> A ⟨ i ⟩v) (vty Γ)) ⟩
    lookup x (vty (Γ ⟪ i ⟫v))
  ∎)
vunit Γ ⟨ i ⟩vv = vunit (Γ ⟪ i ⟫v)
vpair V V' ⟨ i ⟩vv = vpair (V ⟨ i ⟩vv) (V' ⟨ i ⟩vv)
vdestr V Vk ⟨ i ⟩vv = vdestr (V ⟨ i ⟩vv) (Vk ⟨ i ⟩vv)
vthunk M ⟨ i ⟩vv = vthunk (M ⟨ i ⟩cc)
valtrans sq ⟨ Top ⟩vv rewrite ctxttrans-dapp-lem (ValTC.ΓΓ sq) Top = ValTC.V⊤ sq ⟨ Top ⟩vv
valtrans sq ⟨ Bot ⟩vv rewrite ctxttrans-dapp-lem (ValTC.ΓΓ sq) Bot = ValTC.V⊥ sq ⟨ Bot ⟩vv

data Term where
  hole : forall {I}{Γ : VCtx I}{B} -> Term Γ (hstoup B) B
  lambda : forall {I}{Γ : VCtx I}{Δ}{A}{B} ->
           Term (Γ ,, A) Δ B ->
           Term Γ Δ (A ⇀ B)
  app : forall {I}{Γ : VCtx I}{Δ}{A}{B} ->
           Term Γ Δ (A ⇀ B) ->
           Value Γ A ->
           Term Γ Δ B
  force : forall {I}{Γ : VCtx I}{B} ->
          Value Γ (U B) ->
          Term Γ (cdot _) B
  ret : forall {I}{Γ : VCtx I}{A} ->
        Value Γ A ->
        Term Γ (cdot _) (F A)
  lett : forall {I}{Γ : VCtx I}{Δ}{A}{B} ->
         Term Γ Δ (F A) ->
         Term Γ (cdot _) B ->
         Term Γ Δ B

hole ⟨ i ⟩cc = hole
lambda M ⟨ i ⟩cc = lambda (M ⟨ i ⟩cc)
app M V ⟨ i ⟩cc = app (M ⟨ i ⟩cc) (V ⟨ i ⟩vv)
force V ⟨ i ⟩cc = force (V ⟨ i ⟩vv)
ret V ⟨ i ⟩cc = ret (V ⟨ i ⟩vv)
lett M Mk ⟨ i ⟩cc = lett (M ⟨ i ⟩cc) (Mk ⟨ i ⟩cc)

valrefl : forall {Γ}{A} -> Value Γ A -> Value (vctxrefl Γ) (vrefl A)
termrefl : forall {Γ}{Δ}{B} -> Term Γ Δ B -> Term (vctxrefl Γ) (cctxrefl Δ) (crefl B)

valrefl (var Γ x A p) = var (vctxrefl Γ) x (vrefl A) (
  begin
    vrefl A
    ≡⟨ cong vrefl p ⟩
    vrefl (lookup x (vty Γ))
    ≡⟨ sym (lookup-map x vrefl (vty Γ)) ⟩
    lookup x (vty (vctxrefl Γ))
  ∎)
valrefl (vunit Γ) = vunit _
valrefl (vpair V V') = vpair (valrefl V) (valrefl V')
valrefl (vdestr V Vk) = vdestr (valrefl V) (valrefl Vk)
valrefl (vthunk M) = vthunk (termrefl M)

termrefl hole = hole
termrefl (lambda M) = lambda (termrefl M)
termrefl (app M V) = app (termrefl M) (valrefl V)
termrefl (force V) = force (valrefl V)
termrefl (ret V) = ret (valrefl V)
termrefl (lett M Mk) = lett (termrefl M) ((termrefl Mk))
